import pyscratch as pysc
game = pysc.game

spawn_button = pysc.create_single_costume_sprite("assets/fish_red_outline.png")
spawn_button.y = 250

# no guarantee that the `player1` and `player2` are created at this point.
# therefore, do not do this:
#player1 = game['player1']

def spwan_enemy(): 
    enemy_blue = pysc.create_single_costume_sprite("assets/fish_blue_outline.png")

    enemy_blue.y = pysc.random_number(0, game.screen_height)
    enemy_blue.x = 0

    player1 = game['player1']
    player2 = game['player2']
    
    # a better way: break the loop when the sprite is removed
    #while not enemy_blue.removed: 
    while True:
        enemy_blue.x += 4

        if enemy_blue.is_touching(player1):
            enemy_blue.remove()
            player1.scale_by(1.1)

        if enemy_blue.is_touching(player2):
            enemy_blue.remove()
            player2.scale_by(1.1)

        yield 1/game.framerate  # must have an yield in a loop! 
    
spawn_button.when_this_sprite_clicked().add_handler(spwan_enemy) 





import pyscratch as pysc
from pyscratch import game

SCREEN_HEIGHT = 720
SCREEN_WIDTH = 1280

paddle_colour = (200, 200, 200)
paddle_width = 20
paddle_height = 130
paddle_margin = 30


sprite = pysc.create_rect_sprite(paddle_colour, paddle_width, paddle_height,  position=(paddle_margin, SCREEN_HEIGHT//2))
game['left_paddle'] = sprite
sprite.set_draggable(True)

def movement():
    speed = 0
    while True: 
        if pysc.is_key_pressed('w'):
            speed = -8
        if pysc.is_key_pressed('s'):
            speed = 8
        
        speed *= 0.9

        sprite.y += speed
        
        sprite.y = pysc.cap(sprite.y, 0+paddle_height/2, SCREEN_HEIGHT-paddle_height/2)

        yield 1/game.framerate
        
game_start_event = sprite.when_game_start()
game_start_event.add_handler(movement)





import pyscratch as pysc
from pyscratch import game

player = pysc.create_single_costume_sprite("assets/used_by_examples/bullet_hell/player.png")
game['player'] = player
def movement():
    player.set_scale(3)
    player.set_xy((360, 1080))
    

    speed = 8

    while True:
        yield 1/game.framerate

        if pysc.is_key_pressed('w'):
            player.y -= speed

        if pysc.is_key_pressed('s'):
            player.y += speed

        if pysc.is_key_pressed('a'):
            player.x -= speed     

        if pysc.is_key_pressed('d'):
            player.x += speed

player.when_game_start().add_handler(movement)


def shoot_bullet():
    Bullet = game['Bullet']
    bullet_speed = 10
    while True: 
        yield 0.8

        Bullet(player.x, player.y-35, bullet_speed)

player.when_game_start().add_handler(shoot_bullet)


import pyscratch as pysc
import player, player_bullets, enemy

from setting import *

pysc.game.update_screen_mode((SCREEN_WIDTH, SCREEN_HEIGHT))

pysc.create_edge_sprites()
#pysc.game.shared_data['enemy_level'] = 0

pysc.game.start(60, 60, False, event_count=True)


import pyscratch as pysc


# create the sprite and initial settings
#enemy = pysc.create_animated_sprite("assets/other_fishes")
enemy = pysc.create_single_costume_sprite("assets/other_fishes/0.png")


def enemy_on_game_start():
    enemy.set_rotation_style_left_right()
    enemy.hide() # hide the parent

    # clone itself very 2 seconds
    while True: 
        enemy.create_clone()
        yield 2

enemy.when_game_start().add_handler(enemy_on_game_start)


# clone movement
def on_clone(clone_sprite: pysc.Sprite): 


    # random height
    clone_sprite.y = pysc.random_number(0, 720)

    # randomly either from the left or from the right
    if pysc.random_number(0, 1) > 0.5: 
        clone_sprite.x = 0
        clone_sprite.direction = 0 # left to right
    else:
        clone_sprite.x = 1280
        clone_sprite.direction = 180 # right to left

    # random size
    size = pysc.random_number(0.3, 2)
    clone_sprite.set_scale(size)

    # show the sprite
    clone_sprite.show()

    # movement
    while True:
    
        clone_sprite.direction += pysc.random_number(-2, 2)

        clone_sprite.move_indir(2)
        yield 1/60

enemy.when_started_as_clone().add_handler(on_clone)




import pyscratch as pysc


# create the sprite
enemy = pysc.create_single_costume_sprite("assets/other_fishes/0.png")

# Event: when_game_start, clone creation
def enemy_on_game_start():
    enemy.set_rotation_style_left_right()
    enemy.hide() # hide the parent

    # clone itself very 2 seconds
    while True: 
        enemy.create_clone()
        yield 2

enemy.when_game_start().add_handler(enemy_on_game_start)


# Event: when_started_as_clone, clone movement
def clone_movement(clone_sprite):

    screen_height = 720

    # start the fish from the left edge at a random height
    clone_sprite.y = pysc.random_number(0, screen_height)
    clone_sprite.x = 0

    # random size
    size = pysc.random_number(0.8, 1.2)
    clone_sprite.set_scale(size)

    # show the clone
    clone_sprite.show()

    while True:
        
        clone_sprite.move_indir(3)

        yield 1/60

enemy.when_started_as_clone().add_handler(clone_movement)



from shutil import move
import pyscratch as pysc
from pyscratch import game
from pyscratch.game_module import is_key_pressed

player_skin = pysc.create_animated_sprite("assets/used_by_examples/doodle_jump/player")
player = pysc.create_rect_sprite((0,0,0), 70, 20, position=(370, 900))
player.set_transparency(0)


game['player'] = player
game['gravity'] = 500
player['vy'] = 0
player['world_y'] = 900
game['view_y'] = 0

def movement():
    x_speed = 300
    dt = 1/game.framerate
    player_skin.set_rotation_style_left_right()
    #game.bring_to_front(player_skin)
    while True:
        player_skin.x = player.x
        player_skin.y = player.y-60
        yield dt

        player['vy'] += game['gravity']*dt
        player['world_y'] += player['vy']*dt
        player.y = player['world_y']-game['view_y']

        game['view_y'] = min(player['world_y']-900, game['view_y'])

        if pysc.is_key_pressed('left'):
            player.x -= x_speed*dt
            player_skin.direction=0

        if pysc.is_key_pressed('right'):
            player.x += x_speed*dt
            player_skin.direction=180


player.when_game_start().add_handler(movement)


def on_msg_jump(_):
    player['vy'] = -500
    player_skin.set_frame(1)
    yield 0.3
    player_skin.set_frame(0)


player.when_receive_message('jump').add_handler(on_msg_jump)



import pygame

SCREEN_HEIGHT = 720
SCREEN_WIDTH = 1280
DEFAULT_FONT24 = pygame.font.SysFont(None, 24)
DEFAULT_FONT48 = pygame.font.SysFont(None, 24)

LEFT_PANEL_WIDTH = 120
RIGHT_PANEL_WIDTH = 600

BOTTOM_RIGHT_PANEL_HEIGHT = 200
PANEL_MARGIN = 10





import pyscratch as pysc
import pygame
game = pysc.game

text_box = pysc.create_rect_sprite((200, 200, 200, 20), 700, 90)
font = pygame.font.Font('assets/Kenney Future.ttf', size=96)

text_box.write_text("YOU LOSE", font, offset=(350, 45))


text_box.hide()

text_box.set_draggable(True)
text_box.set_xy((260, 195))
text_box.direction = -35



def game_end(): 
    text_box.show()

text_box.when_backdrop_switched(1).add_handler(game_end)




import pyscratch as pysc

player1 = pysc.create_single_costume_sprite("assets/fish_brown_outline.png")
player2 = pysc.create_single_costume_sprite("assets/fish_orange_outline.png")

def move(): 
    while True:
        if pysc.is_key_pressed("d"): 
            player1.x += 4  

        if pysc.is_key_pressed("a"):  
            player1.x -= 4  

        if pysc.is_key_pressed("w"):  
            player1.y -= 4  
            
        if pysc.is_key_pressed("s"):  
            player1.y += 4  

        if pysc.is_key_pressed("right"): 
            player2.x += 4  

        if pysc.is_key_pressed("left"):  
            player2.x -= 4  

        if pysc.is_key_pressed("up"):  
            player2.y -= 4  
            
        if pysc.is_key_pressed("down"):  
            player2.y += 4  

        yield 1/60  # must have an yield in a loop! 

# for the purpose of this tutorial, 
# `player1.when_game_start()` and `player2.when_game_start()` are almost the same. 
game_start = player2.when_game_start()  
game_start.add_handler(move) 




def move2(): 
    player1.x = 200
    player1.y = 400
    
    player2.x = 400
    player2.y = 200
    
    while True:
        if player1.is_touching(player2):
            player1.hide()
            player2.hide()

        yield 1/60  # must have an yield in a loop! 

# for the purpose of this tutorial, 
# `player1.when_game_start()` and `player2.when_game_start()` are almost the same. 
player2.when_game_start().add_handler(move2) 


import pygame
import numpy as np
from settings import *
import pyscratch as pysc
game = pysc.game 
xpos2 = 1050


w, h = 50*1.62, 40
colour = (255, 127, 127)
button = pysc.create_rect_sprite(colour, w, h)
button.set_xy((1200, 650))
button.write_text("Fit", DEFAULT_FONT24, offset=(w/2, h/2))
pysc.game.change_layer(button, 1)



def on_click():
    if not 'image_on_right_display' in pysc.game.shared_data:
        pysc.game.broadcast_message('warning', 'image not selected' )
        return 
            
    if game['offset_x'] is None: return 
    if game['offset_y'] is None: return 
    if game['size_x'] is None: return 
    if game['size_y'] is None: return 


    img: pygame.Surface = game['image_on_right_display']


    fitter = ParamsFitter(img)

    n_col, pixel_x, offset_x = fitter.find_best_x(game['offset_x'], game['limit_x'])
    n_row, pixel_y, offset_y = fitter.find_best_y(game['offset_y'], game['limit_y'])

    if (n_col and pixel_x and (not offset_x is None)):
        game['n_col'] = n_col
        game['pixel_x'] = pixel_x
        game['offset_x'] = offset_x
        game['size_x'] = min(game['limit_x'] - game['offset_x']+1, game['size_x'])

    if (n_row and pixel_y and (not offset_y is None)):
        game['n_row'] = n_row
        game['pixel_y'] = pixel_y
        game['offset_y'] = offset_y
        game['offset_y'] = offset_y
        game['size_y'] = min(game['limit_y'] - game['offset_y']+1, game['size_y'])


    #game.broadcast_message('n_row_change')
    #game.broadcast_message('n_col_change')

    #game.broadcast_message('pixel_x_change')

    #game.broadcast_message('pixel_y_change')

    #game.broadcast_message('offset_x_change')
    #game.broadcast_message('offset_y_change')

    


button.when_this_sprite_clicked().add_handler(on_click)

from itertools import product
class ParamsFitter:

    def __init__(self, surface):

        self.surface = surface
        self.arr = pygame.surfarray.pixels_alpha(surface)

        self.yslits = ParamsFitter.get_slit_values(self.arr.mean(0))
        self.xslits = ParamsFitter.get_slit_values(self.arr.mean(1))
    
    def num_frames_x(self, offset, limit):
        # TODO: think about it: is the boundary correct?
        return ParamsFitter.get_num_frames(self.xslits[offset: limit+2])
    
    def num_frames_y(self, offset, limit):
        return ParamsFitter.get_num_frames(self.yslits[offset: limit+2])
    
    def find_best_x(self, offset_centre: int, limit:int):
        offset_centre = int(offset_centre)
        limit = int(limit)

        return ParamsFitter.search_best(self.xslits[: limit+2], offset_centre)
    
    def find_best_y(self, offset_centre: int, limit:int):
        offset_centre = int(offset_centre)
        limit = int(limit)
        return ParamsFitter.search_best(self.yslits[: limit+2], offset_centre)
    
    @staticmethod
    def search_best(slit_values, centre_offset, step_range_ratio=0.1, offset_range_ratio=0.1):
        #print(len(slit_values))
        n_frame = int(ParamsFitter.get_num_frames(slit_values[centre_offset:]))
        print(ParamsFitter.get_num_frames_rank(slit_values[centre_offset:]))

        # TODO: possibly DIV BY ZERO?
        step_size = max(1, (len(slit_values[centre_offset:])-1)/n_frame )
        
        step_range = round(step_size*step_range_ratio)
        offset_range = round(step_size*offset_range_ratio)

        step_size = round(step_size)

        # TODO: very easy to go out of boundary
        step_sizes = range(step_size+step_range+1, max(0, step_size-step_range), -1)
        offsets = range(max(0,centre_offset-offset_range), centre_offset+offset_range+1)

        vmin = np.inf
        best_ss = None
        best_os = None
        for ss, os in product(step_sizes, offsets):

            n, v = ParamsFitter.eval_cut(slit_values, ss, os)
            #print(f"trying: pixel_sz:{ss}, offset:{os} -> {n, v}")
            if not n == n_frame: continue
            if v < vmin:
                best_ss = ss
                best_os = os
                vmin = v
            #print(ss, os, v)



        print(f"n_frame: {n_frame}, pixel_sz:{best_ss}, offset:{best_os} -> {vmin}")
        return n_frame, best_ss, best_os
        
    @staticmethod
    def get_num_frames(slit_values):
        return np.argmax(np.abs(np.fft.fft(slit_values))[1:len(slit_values)//2])+1 
        
    @staticmethod
    def get_num_frames_rank(slit_values):
        return np.argsort(np.abs(np.fft.fft(slit_values))[1:len(slit_values)//2])+1 
    

    @staticmethod
    def get_rough_est_num_pixel(slit_values):
        return (len(slit_values)-1)/ParamsFitter.get_num_frames(slit_values)

    @staticmethod
    def eval_cut(slit_values, step_size, offset=0):
        cut_indices = np.arange(offset, len(slit_values), step_size)
        return len(cut_indices)-1, slit_values[cut_indices].mean()   
 
    @staticmethod
    def get_slit_values(axis_mean):
        return np.r_[[0], (axis_mean[1:]+axis_mean[:-1]), [0]]



"""
Everything in this module is directly under the pyscratch namespace. 
For example, instead of doing `pysc.helper.random_number`,
you can also directly do `pysc.random_number`
"""


from functools import cache
from itertools import product
from pathlib import Path
from typing import Dict, List, Tuple, Union
import random

import pygame
import numpy as np


def cap(v:float, min_v:float, max_v:float)->float:
    """
    Cap a value to a range between `min_v` and `max_v`.

    Example
    ```python
    cap(5, 1, 10) # returns 5 
    cap(15, 1, 10) # returns 10
    cap(-10, 1, 10) # returns 1
    
    # a more realistic example usage: keep the sprite within the screen. 
    my_sprite.x = cap(my_sprite.x, 0, SCREEN_WIDTH)
    my_sprite.y = cap(my_sprite.y, 0, SCREEN_HEIGHT)
    ```
    """
    return max(min(max_v, v), min_v)

def random_number(min_v:float, max_v:float) -> float:
    """
    Returns a random number between `min_v` and `max_v`
    """
    return random.random()*(max_v-min_v)+min_v

@cache
def load_image(path: str) -> pygame.Surface:
    """
    Return the `pygame.Surface` given the path to an image. 
    """
    return pygame.image.load(path).convert_alpha()


def get_frame_from_sprite_sheet(
        sheet: pygame.Surface, 
        columns:int, rows:int, index:int, 
        spacing:int=0, margin:int=0, inset:int=0):
    """
    Extract a specific frame from a sprite sheet. 
    
    *You will not need to use this function directly.*

    THIS FUNCTION IS WRITTEN BY AN AI. 


    The index starts from the top-left as 0,
    and increments left-to-right before up-to-down.
    For example, for a 3x3 sheet, the index would be
    ```
    0 1 2
    3 4 5
    6 7 8
    ```
    """
    sheet_rect = sheet.get_rect()

    total_spacing_x = spacing * (columns - 1)
    total_spacing_y = spacing * (rows - 1)

    total_margin_x = margin * 2
    total_margin_y = margin * 2

    frame_width = (sheet_rect.width - total_spacing_x - total_margin_x) // columns
    frame_height = (sheet_rect.height - total_spacing_y - total_margin_y) // rows

    col = index % columns
    row = index // columns

    x = margin + col * (frame_width + spacing)
    y = margin + row * (frame_height + spacing)

    # Apply internal cropping (inset) from all sides
    cropped_rect = pygame.Rect(
        x + inset,
        y + inset,
        frame_width - 2 * inset,
        frame_height - 2 * inset
    )

    return sheet.subsurface(cropped_rect)

def get_frame_from_sprite_sheet_by_frame_size(
        sheet: pygame.Surface, 
        size_x: int, size_y: int,
        c:int, r: int):
    """
    Extract a specific frame from a sprite sheet.
    
    *You will not need to use this function directly.*


    Parameters
    ---
    sheet: pygame.Surface
        The sprite sheet you want to cut
    
    size_x: int
        The x pixel size of the frame
    
    size_y: int
        The y pixel size of the frame

    c: int 
        The column position of the frame

    r: int 
        The row position of the frame
    """
    crop_rect = pygame.Rect(
        c*size_x,
        r*size_y,
        size_x,
        size_y
    )

    return sheet.subsurface(crop_rect)

from PIL import Image, ImageSequence
import pygame

@cache
def load_gif_frames(path) -> List[pygame.Surface]:
    """
    Load a gif file as a list of images (as `pygame.Surface`)

    *You will not need to use this function directly.*

    THIS FUNCTION IS WRITTEN BY AN AI.

    """
    pil = Image.open(path)
    frames = []
    for frame in ImageSequence.Iterator(pil):
        mode   = frame.convert("RGBA")
        data   = mode.tobytes()
        size   = mode.size
        surf   = pygame.image.frombuffer(data, size, "RGBA").convert_alpha()
        frames.append(surf)
    return frames


@cache
def load_frames_from_gif_folder(folder_path: Path):
    """
    Creates a `frame_dict` that the Sprite constructor takes using 
    the gif files inside a folder. 

    The folder should be organised in the following way: 
    ```
    ├─ player/
        ├─ walking.gif
        ├─ idling.gif
        ├─ jumping.gif
        ...
    ```

    *You will not need to use this function directly.*


    """
    frame_dict = {}
    for f in folder_path.iterdir():
        if f.suffix != '.gif': 
            print(f"ignoring {f.name} (not a gif)")
            continue

        frame_dict[f.stem] = load_gif_frames(f)
    
    return frame_dict




def cut_sprite_sheet(
        sheet: pygame.Surface, 
        columns, rows, 
        spacing=0, margin=0, inset=0, 
        folder_path='.', 
        suffix='png'):
    """
    Split a sprite sheet so each frame is its own image file. 
    
    *You probably won't need to use this function directly.*
    
    Parameters
    ---
    sheet: pygame.Surface
        The sprite sheet image
    columns: int
        The total number of columns of the sheet
    rows: int
        The total number of rows of the sheet
    spacing: int
        Keep it as zero.
    margin: int
        Keep it as zero.
    inset: int
        Keep it as zero.
    folder_path: str
        The path of the destination folder to put the splitted images. 
    suffix: str
        The file extension 
    """
    
    folder_path = Path(folder_path)
    if not folder_path.exists():
        folder_path.mkdir()
    for i in range(columns*rows):
        f = get_frame_from_sprite_sheet(sheet, columns, rows, i, spacing, margin, inset)
        pygame.image.save(f, folder_path/f"{i}.{suffix}")

def _get_frame_sequence(sheet:pygame.Surface, columns:int, rows:int, indices:List[int], spacing:int, margin:int, inset:int):
    """
    **You will not need to use this function directly.**
    """
    return [get_frame_from_sprite_sheet(sheet, columns, rows, i, spacing, margin, inset) for i in indices]

def _get_frame_dict(
        sheet:pygame.Surface, columns:int, rows:int, 
        indices_dict: Dict[str, List[int]], 
        spacing:int=0, margin:int=0, inset:int=0):
    """
    **You will not need to use this function directly.**
    """
    frame_dict = {}
    for k, v in indices_dict.items():
        assert isinstance(v, list) or isinstance(v, tuple)

        frame_dict[k] = _get_frame_sequence(sheet, columns, rows, v, spacing, margin, inset)

    return frame_dict

@cache
def load_frames_from_folder(folder_path: Union[Path, str]):
    """
    Creates a `frame_dict` that the Sprite constructor takes using 
    the images inside a folder. 
    You can use `sprite.create_animated_sprite` instead for convenience.

    The folder should be organised in one of these two ways: 

    **Option 1**: Only one frame mode. *Note that the images must be numbered.*
    ```
    ├─ player/
        ├─ 0.png
        ├─ 1.png
        ├─ 2.png
        ...
    ```

    **Option 2**: Multiple frame modes. *Note that the images must be numbered.*
    ```
    ├─ player/  
        ├─ walking/  
            ├─ 0.png  
            ├─ 1.png  
            ...
        ├─ idling/
            ├─ 0.png
            ├─ 1.png
            ... 
        ...
    ``` 
    Example
    ```python
    frame_dict = load_frames_from_folder("assets/player")
    my_sprite1 = Sprite(frame_dict)
    ```
    """
    return _load_frames_from_folder_uncached(folder_path)



def _load_frames_from_folder_uncached(folder_path: Union[Path, str]):
    
    def extract_images(path: Path):
        index2image: Dict[int, pygame.Surface] = {}

        for f in path.iterdir():
            if f.is_dir():
                continue

            if not f.stem.isdigit(): 
                print(f'skipping: {f.name}')
                continue
            index2image[int(f.stem)] = pygame.image.load(f).convert_alpha()
        
        return [index2image[i] for i in sorted(index2image.keys())]
        

    path = Path(folder_path)

    folder_seen = False
    file_seen = False

    frame_dict = {}
    for f in path.iterdir():
        if f.is_dir():
            folder_seen = True
            assert not file_seen
            frame_dict[f.stem] = extract_images(f)
        else:
            file_seen = True
            assert not folder_seen

    if not folder_seen:
        frame_dict[path.stem] = extract_images(path)

    return frame_dict

#RGBType = Tuple[int, int, int]
#RGBAType = Tuple[int, int, int, int]
#ColourType = Union[RGBAType, RGBType]

def create_circle(colour, radius: float) -> pygame.Surface:
    """
    Create a circle image (pygame.Surface) given the colour and radius. 

    Parameters
    ---
    colour: Tuple[int, int, int] or Tuple[int, int, int, int]
        The colour of the rectangle in RGB or RGBA. Value range: [0-255].
    radius: float
        the radius of the cirlce.
    """
    surface = pygame.Surface((radius*2, radius*2), pygame.SRCALPHA)
    pygame.draw.circle(surface, colour, (radius, radius), radius)
    return surface


def create_rect(colour, width: float, height: float) -> pygame.Surface:
    """
    Create a rectangle image (pygame.Surface) given the colour, width and height

    Parameters
    ---
    colour: Tuple[int, int, int] or Tuple[int, int, int, int]
        The colour of the rectangle in RGB or RGBA. Value range: [0-255].
    width: float
        the width (x length) of the rectangle.
    height: float
        the height (y length) of the rectangle.
    """

    surface = pygame.Surface((width, height), pygame.SRCALPHA)
    pygame.draw.rect(surface, colour, surface.get_rect())
    return surface

def scale_and_tile(image, screen_size, scale_factor):
    """
    Scale an image by a factor and tile it to fill the screen.
    
    THIS FUNCTION IS WRITTEN BY AN AI. 

    Parameters
    ----------
    image : pygame.Surface
        The image to scale and tile.
    screen_size : tuple of int
        The width and height of the target surface.
    scale_factor : float
        Factor by which to scale the image.
    """

    img_w, img_h = image.get_size()
    new_img = pygame.transform.smoothscale(image, (int(img_w * scale_factor), int(img_h * scale_factor)))
    new_img_w, new_img_h = new_img.get_size()
    
    tiled_surface = pygame.Surface(screen_size)
    
    for y in range(0, screen_size[1], new_img_h):
        for x in range(0, screen_size[0], new_img_w):
            tiled_surface.blit(new_img, (x, y))
    
    return tiled_surface


def scale_to_fill(image:pygame.Surface, new_size:Tuple[int, int]):
    """
    Scale an image to a new size without preserving aspect ratio.

    Parameters
    ----------
    image : pygame.Surface
        The image to scale.
    new_size : tuple of int
        Target width and height of the image.
    """
    return pygame.transform.smoothscale(image, new_size)


def scale_to_fit_aspect(image:pygame.Surface, new_size:Tuple[int, int], fit='horizontal'):
    """
    Scale an image to fit a rect while preserving aspect ratio.
    
    THIS FUNCTION IS WRITTEN BY AN AI. 

    Parameters
    ----------
    image : pygame.Surface
        The image to scale.
    new_size : tuple of int
        The width and height of the target rect.
    fit : 'horizontal' or 'vertical'
        Axis to fit the image against. Default is 'horizontal'.
    """
    img_rect = image.get_rect()
    screen_w, screen_h = new_size
    img_w, img_h = img_rect.size

    if fit == 'horizontal':
        scale_factor = screen_w / img_w
    elif fit == 'vertical':
        scale_factor = screen_h / img_h
    else:
        raise ValueError("fit must be either 'horizontal' or 'vertical'")

    new_size = (int(img_w * scale_factor), int(img_h * scale_factor))
    return pygame.transform.smoothscale(image, new_size)


def _set_transparency(image, factor):
    """
    Set the transparency of an image.

    THIS FUNCTION IS WRITTEN BY AN AI. 

    ***IMCOMPLETE IMPLEMENTATION***: 
    The transparency of the transparent background of the image is also changed


    Parameters
    ----------
    image : pygame.Surface
        The image to adjust.
    factor : float
        Transparency level from 0.0 (fully transparent) to 1.0 (fully opaque).
    """
    new_image = image.copy()
    new_image.set_alpha(int(factor*255))
    return new_image

def set_transparency(image: pygame.Surface, factor):
    """
    Get a new copy of the image with the transparency applied.

    Parameters
    ----------
    image : pygame.Surface
        The image to adjust.
    factor : float
        Transparency level from 0.0 (fully transparent) to 1.0 (fully opaque).
    """
    w = image.get_width()
    h = image.get_height()

    sur = pygame.Surface((w, h)).convert_alpha()

    sur.fill((255,255,255,int(factor*255) ))
    #sur.set_alpha(int(factor*255))

    new_image = image.copy()
    new_image.blit(sur, (0,0), special_flags=pygame.BLEND_RGBA_MULT)
    return new_image


def adjust_brightness(image, factor):
    """
    Get a new copy of the image with the brightness changes applied.
    
    THIS FUNCTION IS WRITTEN BY AN AI. 

    Parameters
    ----------
    image : pygame.Surface
        The image to adjust.
    factor : float
        Brightness multiplier. Values < 1.0 darken, values > 1.0 brighten.
       
    """
    new_image = image.copy()
    brightness_surface = pygame.Surface(image.get_size()).convert_alpha()
    brightness_surface.fill((255, 255, 255, 0))  # Start with no change
    
    # Per-pixel operation is slow. Instead, we use special_flags to modulate brightness.
    if factor < 1:
        # Darken using multiply blend mode
        darken_surface = pygame.Surface(image.get_size()).convert_alpha()
        value = int(255 * factor)
        darken_surface.fill((value, value, value))
        new_image.blit(darken_surface, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)
    elif factor > 1:
        # Brighten by blending in white
        value = int(255 * (factor - 1))
        brighten_surface = pygame.Surface(image.get_size()).convert_alpha()
        brighten_surface.fill((value, value, value))
        new_image.blit(brighten_surface, (0, 0), special_flags=pygame.BLEND_RGBA_ADD)

    return new_image


def _draw_guide_lines(screen: pygame.Surface, font: pygame.font.Font, minor_grid, major_grid):
    w, h = screen.get_width(), screen.get_height()
    

    pygame.draw.lines(screen, (0,0,0), True, [(0,0), (0, h), (w, h), (w, 0)])
    screen.blit(font.render(f"({0},{0})", True, (0,0,0)), (5,5))
    for i in range(0, w, minor_grid):
        lw = 1
        v = 200
        if i and (not i % major_grid):
            lw = 1
            v = 0
            screen.blit(font.render(f"{i}", True, (0,0,0)), (i+5,5))

        pygame.draw.line(screen, (v,v,v,255-v), (i, 0), (i, h), width=lw)


        

    for i in range(0, h, minor_grid):
        lw = 1
        v = 200
        if i and (not i % major_grid):
            lw = 1
            v = 0
            screen.blit(font.render(f"{i}", True, (0,0,0)), (5,i+5))        
        pygame.draw.line(screen, (v,v,v,255-v), (0, i), (w, i), width=lw)

    #text = font.render(f"({w},{h})", True, (0,0,0))
    #screen.blit(text, (w-5-text.get_width(), h-5-text.get_height()))


    for x, y in product(range(major_grid, w, major_grid), range(major_grid, h, major_grid)):
        text = font.render(f"({x},{y})", True, (0,0,0))
        screen.blit(text, (x+5, y+5))



def _show_mouse_position(screen, font):
    w, h = screen.get_width(), screen.get_height()
    
    mx, my = pygame.mouse.get_pos()

    text = font.render(f"({mx},{my})", True, (0,0,0))
    screen.blit(text, (w-5-text.get_width(), h-5-text.get_height()))


import pyscratch as pysc


player = pysc.create_single_costume_sprite("assets/player-fish.png")
# pysc.game['player'] = player  # put player as a shared variable so other sprites can access it

# Event: game start, the movement of the player
def on_game_start():
      
    player.set_rotation_style_left_right()
    speed_decay = 0.9
    speed_y = 0
    speed_x = 0

    while True:
        max_speed = 4


        if pysc.is_key_pressed('w'):
            speed_y = -max_speed

        elif pysc.is_key_pressed('s'):
            speed_y = max_speed

        else:
            speed_y *= speed_decay # speed_y = speed_y * speed_decay

        if pysc.is_key_pressed('a'):
            player.direction = 180
            speed_x = -max_speed
            
        elif pysc.is_key_pressed('d'):
            player.direction = 0
            speed_x = max_speed
        else:
            speed_x *= speed_decay # speed_x = speed_y * speed_decay


        player.y += speed_y
        player.x += speed_x

        yield 1/60

# or shorter: player.when_game_start().add_handler(on_game_start)
game_start_event = player.when_game_start()
game_start_event.add_handler(on_game_start)



# Event: game start, ocean current changes
pysc.game['current_x'] = 0
pysc.game['current_y'] = 0

def ocean_current_change():
    # slowly change the current variables every 0.5 second
    while True:
        pysc.game['current_x'] += pysc.random_number(-0.1, 0.1)
        pysc.game['current_y'] += pysc.random_number(-0.1, 0.1)
        yield 0.5

pysc.game.when_game_start().add_handler(ocean_current_change)

# option 2: almost the same as above but the event will be removed when player is removed.
# player.when_game_start().add_handler(ocean_current_change)




# Event: game start, moved by the ocean current
def ocean_current_movement():
    while True:

        # you can add these two lines inside the loop in the first event as well
        # but for the code readabilty, 
        # it can be a good idea to put it out as a separate event. 
        player.x += pysc.game['current_x'] 
        player.y += pysc.game['current_y'] 
        yield 1/60

player.when_game_start().add_handler(ocean_current_movement)


"""
broadcast messages: 
    order_change ((old, new))
    preview_click(dict(surface=surface, order=order))
    change_animation_done (frames: List)

    
receice messages:
    change_animation
        reload the animation from the animation path
        remove and create the frame cards for this animation

shared_data:
    scrolling_offset (read only)
    animation_path 
    frame_card_list 
"""

from typing import Dict, List

import numpy as np
import pyscratch as pysc
from pyscratch.sprite import Sprite
from settings import *
from pygame import Surface
from math import floor
from pathlib import Path


# Sprite Function: the frame preview card
FRAME_PREVIEW_MARGIN = 10
preview_width = LEFT_PANEL_WIDTH-FRAME_PREVIEW_MARGIN*2
preview_height = preview_width 

def FramePreviewCard(surface:Surface, order):

    preview_bg = pysc.create_rect_sprite(
        (255, 255, 255),
        preview_width, 
        preview_height
    )
    preview_bg['surface'] = surface
    preview_bg['order'] = order
    preview_bg.oob_limit = np.inf
    
    preview_bg.set_draggable(True)
    

    # set xy
    ypos = FRAME_PREVIEW_MARGIN + order*(preview_height+FRAME_PREVIEW_MARGIN) + preview_height/2
    
    xpos = LEFT_PANEL_WIDTH/2
    preview_bg.set_xy((xpos, ypos))
    preview_bg.y += pysc.game['scrolling_offset']

    # display preview
    image_margin = 20
    fit = "horizontal" if surface.get_width() >= surface.get_height() else "vertical"
    pv_surface = pysc.scale_to_fit_aspect(surface, (preview_width-image_margin, preview_height-image_margin ), fit)
    preview_bg.draw(pv_surface, (preview_width/2, preview_height/2))
    

    # send message on click
    def on_click():


        pysc.game.broadcast_message('preview_click', dict(surface=surface, order=order))
        pysc.game.broadcast_message('show_bin', None)


    preview_bg.when_this_sprite_clicked().add_handler(on_click)


    # scrolling
    def on_scrolling(offset):
        preview_bg.y = ypos + offset
    preview_bg.when_receive_message('scrolling').add_handler(on_scrolling)


        

    # dragging
    def on_mouse_release():
        pysc.game.broadcast_message('hide_bin', None)
        
        left_panel: pysc.Sprite = pysc.game['left_panel']  

        frame_card_list = pysc.game['frame_card_list']



        new_order = ((preview_bg.y-pysc.game['scrolling_offset'])-preview_height/2-FRAME_PREVIEW_MARGIN)/(preview_height+FRAME_PREVIEW_MARGIN)        
        new_order = floor(new_order+0.5)


        preview_bg.x = xpos

        preview_bg.y = ypos + pysc.game['scrolling_offset']

        new_order = min(max(new_order, 0), len(frame_card_list)-1) 

        if not (to_remove := pysc.game['frame_bin'].is_touching_mouse()):
            if not left_panel.is_touching(preview_bg): return
            if new_order == order: return 
            

        print(f"{order} -> {new_order}")

        first_half = frame_card_list[:order]
        second_half = frame_card_list[order+1:] if (order+1)<len(frame_card_list) else []

        temp = first_half + second_half

        #print([c['order'] for c in temp] )

        if not to_remove: 
            temp = temp[:new_order] + [preview_bg] + (temp[new_order:] if (new_order)<len(temp) else [])

        #print([c['order'] for c in temp] )

        write_new_frame_list([f['surface'] for f in temp])

        # update 
        pysc.game.broadcast_message('change_animation', pysc.game['animation_path'])
            
    pysc.game.when_this_sprite_click_released(preview_bg).add_handler(on_mouse_release)

    return preview_bg


def on_cut_sprite_frame_drop(data):
    if not 'animation_path' in pysc.game.shared_data:
        pysc.game.broadcast_message('warning', "Animation not selected yet.")
        return



    sprite: pysc.Sprite = data['sprite']
    surface: Surface = data['surface']
    mos_x, mos_y = data['position']
    frame_card_list = pysc.game['frame_card_list']


    left_panel: pysc.Sprite = pysc.game['left_panel']  
    
    if not left_panel.is_touching_mouse(): return

    

    order = ((mos_y-pysc.game['scrolling_offset'])-preview_height/2-FRAME_PREVIEW_MARGIN)/(preview_height+FRAME_PREVIEW_MARGIN)        
    order = floor(order+0.5)



    frame_card_list = pysc.game['frame_card_list']

    first_half = frame_card_list[:order]
    second_half = frame_card_list[order:] if (order)<len(frame_card_list) else []

    frame_list = [f['surface'] for f in first_half] +[surface]+ [f['surface'] for f in second_half]

    write_new_frame_list(frame_list)

    # update 
    pysc.game.broadcast_message('change_animation', pysc.game['animation_path'])

    sprite.x = sprite['x']
    sprite.y = sprite['y']


pysc.game.when_receive_message('cut_sprite_frame_drop').add_handler(on_cut_sprite_frame_drop)




# helper function
import shutil
def write_new_frame_list(frame_list: List[Surface]):
    # TODO: frame quality degrade over time?
    temp_folder = Path('temp')
    temp_folder.mkdir(exist_ok=True)
    shutil.rmtree(temp_folder)
    temp_folder.mkdir()

    suffix = 'png' # TODO: allow changing suffix 
    for i, f in enumerate(frame_list):
        pygame.image.save(f, temp_folder/f"{i}.{suffix}")

    shutil.rmtree(pysc.game['animation_path'])
    shutil.copytree(temp_folder, pysc.game['animation_path'])
    shutil.rmtree(temp_folder)
    


# event: when there is a change to the animation
# reload the animation from the animation path
pysc.game['frame_card_list'] = []
def on_change_animation(path: Path):

    #pysc.game['scrolling_offset'] = 0



    pysc.game['animation_path'] = path
    for c in pysc.game['frame_card_list']:
        c.remove()
    pysc.game['frame_card_list'] = []

    def extract_images(path: Path):
        index2image: Dict[int, pygame.Surface] = {}

        for f in path.iterdir():
            if f.is_dir():
                continue

            if not f.stem.isdigit(): 
                print(f'skipping: {f.name}')
                continue
            index2image[int(f.stem)] = pygame.image.load(f).convert_alpha()
        
        return [index2image[i] for i in sorted(index2image.keys())]
        
    try: 
        frames = extract_images(path)
    except:
        print('invalid folder structure.')
        return 
    

    for i, f in enumerate(frames):
        pysc.game.shared_data['frame_card_list'].append(
            FramePreviewCard(f, i)
        )
        
    pysc.game.broadcast_message("change_animation_done", frames)
            
    
pysc.game.when_receive_message('change_animation').add_handler(on_change_animation)


import pyscratch as pysc
from pyscratch import game

game['player_bullets'] = []
def Bullet(x, y, speed):
    bullet = pysc.create_single_costume_sprite(
        "assets/used_by_examples/bullet_hell/orb_bullets/6.png",
        position=(x,y)
    )
    bullet.x = x
    bullet.y = y
    bullet.set_scale(0.5)

    def movement(_):
        while True:
            yield 1/game.framerate
            bullet.y -= speed

    bullet.when_timer_above(0).add_handler(movement)


    game['player_bullets'].append(bullet)
    #return bullet

game['Bullet'] = Bullet


def clear_removed_bullet(_):
    game['player_bullets'] = [b for b in game['player_bullets'] if not b.removed]

game.when_timer_reset(2).add_handler(clear_removed_bullet)


from utils.render_wrapped_file_name import render_wrapped_file_name
import pyscratch as pysc
from settings import *

w, h = 400, 100
message_display = pysc.create_rect_sprite((255, 255, 255), w, h)
message_display.hide()
ori_xy = 200, 200
message_display.set_draggable(True)

def on_warning_message(message):

    message_display.show()
    message_display.set_xy(ori_xy)
    txt = render_wrapped_file_name(message, 200, DEFAULT_FONT24, color=(255,0,0))
    message_display.draw(txt, offset=(w/2, h/2))



message_display.when_receive_message('warning').add_handler(on_warning_message)

def on_click():
    message_display.hide()

message_display.when_this_sprite_clicked().add_handler(on_click)



import pyscratch as pysc
import player # Very important!
import enemy # Very important!
import enemy_red
import pygame
# start the game
WIN_WIDTH = 720 # change me
WIN_HEIGHT = 720 # change me
framerate = 60

pysc.game.update_screen_mode((WIN_WIDTH, WIN_HEIGHT)) 
pysc.game.start(framerate, use_frame_time=True)



from pathlib import Path
import pyscratch as pysc
from settings import *

from right_panel import file_display_area, spritesheet_view, cut_parameters, back_button, cut_button, ss_select_corner, cut_parameter_fitting
from main_panel import sprite_edit_ui, play_edit_ui, animation_display, frame_bin, set_as_sprite_folder, warning_message
from left_panel import frame_preview_card, frame_preview_panel

bg = pysc.create_rect((221, 221, 221), SCREEN_WIDTH, SCREEN_HEIGHT)
pysc.game.update_screen_mode((SCREEN_WIDTH, SCREEN_HEIGHT))
pysc.game.set_backdrops([bg])
pysc.game.switch_backdrop(0)

pysc.game.broadcast_message('folder_update', Path('assets'))
pysc.game.broadcast_message('cut_or_nav_mode_change', 'nav')
#pysc.game.broadcast_message('change_sprite_selection', Path('./testing_assets/cat'))

pysc.game.start(30)


import pyscratch as pysc
import random
from setting import *
from enemy_bullets import *
def spawn_line():
    n = 10
    margin = 20
    itv = (SCREEN_WIDTH-margin*2)//(n-1)
    y=0
    for i in range(n):

        x = itv*i+margin

        create_standard_enemy((x, y), 90, False, False, 5, 1000)


def spawn_6_side_entry(_):

    pos1 = (0, 200)
    pos2 = (0, 400)
    pos3 = (0, 600)

    pos4 = (SCREEN_WIDTH, 200)
    pos5 = (SCREEN_WIDTH, 400)
    pos6 = (SCREEN_WIDTH, 600)


    create_standard_enemy(pos1, 45, True, False, 5)
    create_standard_enemy(pos2, 45, True, False, 5)
    create_standard_enemy(pos3, 45, True, False, 5)


    create_standard_enemy(pos4, 90+45, True, False, 5)
    create_standard_enemy(pos5, 90+45, True, False, 5)
    create_standard_enemy(pos6, 90+45,True,  False, 5)






def spawn_random(_):
    pos = pysc.helper.random_number(0, SCREEN_WIDTH), 0
    speed = pysc.helper.random_number(9, 10)
    rotation = pysc.helper.random_number(90-10, 90+10)

    create_standard_enemy(pos, rotation, False, False, speed)
    

def spawn_kamikaze(_):
    pos = pysc.helper.random_number(0, SCREEN_WIDTH), 0
    speed = pysc.helper.random_number(7, 12)

    create_standard_enemy(pos, 90, False, True, speed, bullet_period=1000000)
    


def create_standard_enemy(position, rotation, start_point, pointing_to_player, speed, bullet_period=150):

    

    # create the sprite
    enemy_sprite = pysc.create_rect_sprite((255, 0, 0), 50, 30, pos=position)
    #pysc.game.add_sprite(enemy_sprite)
    enemy_sprite.set_collision_type(ENEMY_TYPE)

    enemy_sprite.add_rotation(rotation)

    if start_point:
        enemy_sprite.point_towards_sprite(pysc.game.shared_data['player'])
        

    # behaviour
    ## 1. move a straight line
    ## 2. shoot bullets (of type based on the level) at a constant interval (based on the level)
    ## 3. hitting a player 
    ## 4. hit by player buller
    ## 5. destroyed when leaving the screen


    ## 1. move a straight line   
    movement_event = enemy_sprite.when_timer_reset(20)
    movement_event.add_handler(lambda x: enemy_sprite.move_indir(speed))

    if pointing_to_player: 
        movement_event.add_handler(
            lambda x: enemy_sprite.point_towards_sprite(pysc.game.shared_data['player'])
        )


    ## 2. shoot bullets (of type based on the level) at a constant interval (based on the level)
    bullet_event = enemy_sprite.when_timer_reset(bullet_period).add_handler(lambda x: create_straight_bullet((enemy_sprite.x, enemy_sprite.y), enemy_sprite.get_rotation()))


    ## 3. hitting a player
    when_hit_player = pysc.game.when_condition_met(lambda: game.is_touching(pysc.game, enemy_sprite, pysc.game.shared_data['player']), repeats=1, associated_sprites=[enemy_sprite])
    when_hit_player.add_handler(lambda x: pysc.game.broadcast_message('player_health', -1))

    ## 4. hit by player buller
    when_hit_by_player_bullet = pysc.game._create_type2type_collision_trigger(PLAYER_BULLET_TYPE, ENEMY_TYPE, associated_sprites=[enemy_sprite])
    
    ## 5. destroyed when leaving the screen
    when_leaving_screen =pysc.game.when_condition_met(lambda: (enemy_sprite.y > SCREEN_HEIGHT), repeats=1, associated_sprites=[enemy_sprite])

    ### callback for 3, 4, 5
    def destroy(x):
        #movement_event.remove()
        #bullet_event.remove()
        #when_hit_player.remove()
        #when_hit_by_player_bullet.remove()
        #when_leaving_screen.remove()
        pysc.game._remove_sprite(enemy_sprite)

    def check_collision(a):
        if enemy_sprite._shape in a.shapes:
            destroy(None)
    
    when_hit_by_player_bullet.add_handler(check_collision)
    when_hit_player.add_handler(destroy)
    when_leaving_screen.add_handler(destroy)


# testing only
pysc.game.when_timer_reset(500, 20).add_handler(spawn_random)


pysc.game.when_timer_reset(3000, 1).add_handler(spawn_6_side_entry)
pysc.game.when_timer_reset(2000, 1).add_handler(spawn_line)





import random
import re, sys
import numpy as np
import pymunk
from pyscratch import game
from pyscratch.sprite import Sprite, create_rect, create_rect_sprite
from pyscratch.helper import _get_frame_dict
from pyscratch.game_module import Game

from main import game, frames, WIDTH, HEIGHT, ENEMY_TYPE, PLAYER_BULLET_TYPE


def create_bullet_attracted(position):
    bullet = Sprite(frames, "spin", position)
    
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    speed = random.random()*15+5

    bullet.point_towards_sprite(game.shared_data['player'])


    movement_event = game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    following_event = game.when_timer_reset(200, 5).add_handler(lambda x: bullet.point_towards_sprite(game.shared_data['player']))

    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        following_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)

    when_exit_screen = game.when_condition_met(lambda: bullet.y > HEIGHT, repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)



def create_bullet_start_pointing(position, _):

    bullet = Sprite(frames, "spin", position)

    

    bullet.point_towards_sprite(game.shared_data['player'])
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    speed = random.random()*15+5
    bullet.point_towards_sprite(game.shared_data['player'])


    movement_event = game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    #movement_event = game.create_timer_trigger(20).on_reset(lambda x: bullet.move_across_dir((random.random()-0.5)*speed*0.3))


    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)

    when_exit_screen = game.when_condition_met(lambda: bullet.y > HEIGHT, repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)
    pass




def create_bullet_move_sine(position, rotation):
    bullet = Sprite(frames, "square_bullets", position)
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    bullet.set_rotation(rotation+90)
    #speed = random.random()*15+5
    speed = 10


    bullet.sprite_data['phase'] = 0
    movement_event = game.when_timer_reset(30)
    def move(x):
        bullet.sprite_data['phase']+=.1
        angle = np.tanh(np.cos(.6*bullet.sprite_data['phase']))


        bullet.add_rotation(angle)
                            
        bullet.move_indir(speed)

    
    movement_event.add_handler(move)


    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)

    when_exit_screen = game.when_condition_met(lambda: (bullet.y > HEIGHT) or (bullet.y < 0) or (bullet.x <0) or (bullet.x>WIDTH) , repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)




def create_bullet_type1(position, rotation):
    bullet = Sprite(frames, "spin", position)
    bullet.add_rotation(rotation)
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    speed = random.random()*15+5
    #speed = 15

    movement_event = game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    #movement_event = game.create_timer_trigger(20).on_reset(lambda x: bullet.move_across_dir((random.random()-0.5)*speed*0.3))


    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)

    when_exit_screen = game.when_condition_met(lambda: bullet.y > HEIGHT, repeats=1)

    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)
        when_exit_screen.remove()

    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)






def create_enemy_type1(position):

    enemy_sprite = create_rect_sprite((255, 0, 0), 50, 30, pos=position)
    #game.add_sprite(enemy_sprite)

    enemy_sprite.add_rotation(90+(random.random()-0.5)*15)
    enemy_sprite.set_collision_type(ENEMY_TYPE)


    speed = random.random()*6
    print(game.shared_data['player'])



    movement_event = game.when_timer_reset(20).add_handler(lambda x: enemy_sprite.move_indir(speed))
    bullet_event = game.when_timer_reset(150).add_handler(lambda x: create_bullet_type1((enemy_sprite.x, enemy_sprite.y), enemy_sprite.get_rotation()))
    #bullet_event = game.create_timer_trigger(1500).on_reset(lambda x: create_bullet_attracted((enemy_sprite.x, enemy_sprite.y)))

    when_hit_player = game.when_condition_met(lambda: game.is_touching(game, enemy_sprite, game.shared_data['player']), repeats=1)
    when_leaving_screen = game.when_condition_met(lambda: (enemy_sprite.y > HEIGHT), repeats=1)
    when_hit_by_player_bullet = game._create_type2type_collision_trigger(PLAYER_BULLET_TYPE, ENEMY_TYPE)


    


    def destroy(x):
        when_hit_player.remove()
        when_leaving_screen.remove()
        when_hit_by_player_bullet.remove()
        movement_event.remove()
        bullet_event.remove()
        game._remove_sprite(enemy_sprite)

    def check_collision(a):
        if enemy_sprite._shape in a.shapes:
            destroy(None)

    when_hit_player.add_handler(destroy)
    when_hit_player.add_handler(lambda x: game.broadcast_message('player_health', -1))
    
    when_leaving_screen.add_handler(destroy)

    when_hit_by_player_bullet.add_handler(check_collision)


    




import pyscratch as pysc
game = pysc.game

enemy = pysc.create_single_costume_sprite("assets/fish_red_outline.png")
enemy.set_draggable(True) # optional: make the sprite draggable

def appear():
    enemy.x = game.screen_width/2
    enemy.y = game.screen_height/2
    enemy.hide()
    yield 3
    enemy.show()

enemy.when_game_start().add_handler(appear)


import random
import re, sys
import numpy as np
import pymunk
from pyscratch import game
from pyscratch.sprite import Sprite, create_rect, create_rect_sprite
from pyscratch.helper import _get_frame_dict
from pyscratch.game_module import Game
from main import *
from enemy import *

def game_start(data):

    player = create_rect_sprite((0, 0, 255), 50, 30, pos=(720//2, 1200))
    game._add_sprite(player)
    game.create_edges()
    player.set_collision_type(PLAYER_TYPE)


    healthbar_red = create_rect((255, 0, 0), 60, 50)

    healthbar_empty = create_rect_sprite((255, 255, 255), 60, 5, pos=(0,0))



    game.shared_data['player'] = player
    game.shared_data['inaccuracy'] = 5

    player.sprite_data['health'] = 10


    game._add_sprite(healthbar_empty)
    healthbar_empty.lock_to(player, (0,-30))
    healthbar_empty.blit(healthbar_red, (0,0))


    #game.create_timer_trigger(100, 20).on_reset(lambda x: create_bullet_move_sine((355, 1), 0))
    game.when_timer_reset(100, 20).on_reset(lambda x: create_bullet_attracted((355, 300)))
    #game.create_timer_trigger(100, 100).on_reset(lambda x: create_bullet_type1((100, 100), 90))

    #game.create_timer_trigger(100, 10)
    
    #game.create_timer_trigger(1200).on_reset(lambda x: shoot_player_bullet((player.x, player.y), game.shared_data['inaccuracy']))
    #game.create_timer_trigger(500, 30).on_reset(lambda x: create_enemy_type1((random.random()*WIDTH, 0)))


    def run_forever(_):
        if game.is_key_pressed(['w']):
            player.move_xy((0, -5))

        if game.is_key_pressed(['s']):
            player.move_xy((0, 5))

        if game.is_key_pressed(['a']):
            player.move_xy((-5, 0))

        if game.is_key_pressed(['d']):
            player.move_xy((5, 0))

        player.set_xy((cap(player.x, 50, WIDTH-50), cap(player.y, HEIGHT-500, HEIGHT)))

        
    def on_health_change(change):
        player.sprite_data['health'] += change
        new_health = max(0, player.sprite_data['health'])

        healthbar_red = create_rect((255, 0, 0), 60*(new_health/10), 50)
        healthbar_empty.blit(healthbar_red)



    game.when_timer_reset(1000/120).on_reset(run_forever)
    game.when_receive_message('player_health').add_handler(on_health_change)


game.when_timer_reset(0.1, 1).on_reset(game_start)


import pyscratch as pysc
import player # Very important!

# start the game
WIN_WIDTH = 500 # change me
WIN_HEIGHT = 500 # change me
framerate = 60

pysc.game.update_screen_mode((WIN_WIDTH, WIN_HEIGHT)) 
pysc.game.start(framerate)



from pathlib import Path
from typing import List, Tuple
import pyscratch as pysc
from pyscratch.sprite import Sprite
from settings import *
from .file_display import FileDisplay
w, h = 50*1.62, 50
colour = (127, 127, 127)
button = pysc.create_rect_sprite(colour, w, h)
button.set_xy((1100, 550))
pysc.game.change_layer(button, 1)

def on_click():
    if pysc.game.shared_data['cut_or_nav_mode'] == 'nav':
        #pysc.game.broadcast_message('cut_or_nav_mode_change', 'cut')
        pass
    else:
        pysc.game.broadcast_message('cut', None)
        #pysc.game.broadcast_message('cut_or_nav_mode_change', 'nav')

button.when_this_sprite_clicked().add_handler(on_click)


def on_msg_mode_change(mode):
    pysc.game.shared_data['cut_or_nav_mode'] = mode
    if mode == 'nav':
        button.write_text("...", DEFAULT_FONT24, offset=(w/2, h/2))
    else:
        button.write_text("Cut", DEFAULT_FONT24, offset=(w/2, h/2))


button.when_receive_message('cut_or_nav_mode_change').add_handler(on_msg_mode_change)


import random
import re, sys
import numpy as np
import pymunk
import pyscratch.game_module
from pyscratch.sprite import Sprite, create_rect, create_rect_sprite, create_edge_sprites
from pyscratch.helper import _get_frame_dict
from pyscratch.game_module import Game
import pygame
WIDTH = 720
HEIGHT = 1280


ENEMY_TYPE = 3
PLAYER_TYPE = 2
PLAYER_BULLET_TYPE = 4
EDGE_TYPE = 1

def cap(v, min_v, max_v):
    return max(min(max_v, v), min_v)



game = Game((WIDTH, HEIGHT))
sprite_sheet = pygame.image.load("assets/09493140a07b68502ef63ff423a6da3954d36fd8/Green Effect and Bullet 16x16.png").convert_alpha()

font = pygame.font.SysFont(None, 24)  # None = default font, 48 = font size
game._suppress_type_collision(PLAYER_TYPE, True)


frames = _get_frame_dict(sprite_sheet, 36, 13, {
    "spin": [i+4*36 for i in range(14, 17+1)], 
    "star_explosion": [i+4*36 for i in range(19, 22+1)], 
    "heal": [i+1*36 for i in range(24, 28+1)], 
    "circle_explosion": [i+5*36 for i in range(14, 17+1)], 


    "square_bullets": [i+9*36 for i in range(24, 28+1)]+[i+9*36 for i in range(27, 24, -1)], 
    "circle_bullets": [i+8*36 for i in range(24, 28+1)]+[i+8*36 for i in range(27, 24, -1)], 

    "shield": [i+5*36 for i in [17]], 

    "bullet1": [i+3*36 for i in range(7, 7+1)]
})



start_buttom = create_rect_sprite((200, 0, 0), width=150, height=60, pos=(game._screen.get_width()//2, game._screen.get_height()//2))
start_buttom.write_text("Click to Start", font)
#game.add_sprite(start_buttom)

def on_click():
    start_buttom.scale_by(0.9)

    on_condition = game.when_condition_met(
        lambda: (not pyscratch.game.get_mouse_presses()[0]), repeats=1)
    
    def start_game(x):
        start_buttom.scale_by(1/0.9)
        game.broadcast_message('game_start', {'count': 0})
        game._remove_sprite(start_buttom)

    on_condition.add_handler(start_game)

#game.retrieve_sprite_click_trigger(start_buttom).add_callback(on_click)
on_click()


def shoot_player_bullet(origin, inaccuracy):

    bullet = Sprite(frames, "circle_bullets", origin)

    #game.add_sprite(bullet)
    bullet.set_collision_type(PLAYER_BULLET_TYPE)
    bullet.set_rotation(-90 + inaccuracy*random.random()-inaccuracy/2)

    movement_timer = game.when_timer_reset(1000/240).add_handler(
        lambda x: bullet.move_indir(2)
    )

    next_frame_timer = game.when_timer_reset(100).add_handler(
        lambda x: bullet.next_frame()
    )


    def destroy_when_exit(x):
        movement_timer.remove()
        next_frame_timer.remove()
        game._remove_sprite(bullet)


    destroy_condition = game.when_condition_met(lambda: (bullet.y < 0), repeats=1)
    destroy_condition.add_handler(destroy_when_exit)


def create_bullet_attracted(position):
    bullet = Sprite(frames, "spin", position)
    
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    speed = random.random()*15+5


    movement_event = game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    following_event = game.when_timer_reset(200, 5).add_handler(lambda x: bullet.point_towards_sprite(game.shared_data['player']))

    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        following_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)

    when_exit_screen = game.when_condition_met(lambda: bullet.y > HEIGHT, repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)



def create_bullet_start_pointing(position, _):

    bullet = Sprite(frames, "spin", position)

    

    bullet.point_towards_sprite(game.shared_data['player'])
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    speed = random.random()*15+5
    bullet.point_towards_sprite(game.shared_data['player'])


    movement_event = game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    #movement_event = game.create_timer_trigger(20).on_reset(lambda x: bullet.move_across_dir((random.random()-0.5)*speed*0.3))


    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)

    when_exit_screen = game.when_condition_met(lambda: bullet.y > HEIGHT, repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)
    pass




def create_bullet_move_sine(position, rotation):
    bullet = Sprite(frames, "square_bullets", position)
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    bullet.set_rotation(rotation+90)
    #speed = random.random()*15+5
    speed = 10


    bullet.sprite_data['phase'] = 0
    movement_event = game.when_timer_reset(30)
    def move(x):
        bullet.sprite_data['phase']+=.1
        angle = np.tanh(np.cos(.6*bullet.sprite_data['phase']))


        bullet.add_rotation(angle)
                            
        bullet.move_indir(speed)

    
    movement_event.add_handler(move)


    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)

    when_exit_screen = game.when_condition_met(lambda: (bullet.y > HEIGHT) or (bullet.y < 0) or (bullet.x <0) or (bullet.x>WIDTH) , repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)




def create_bullet_type1(position, rotation):
    bullet = Sprite(frames, "spin", position)
    bullet.add_rotation(rotation)
    bullet.set_scale(2.3)
    #game.add_sprite(bullet)
    speed = random.random()*15+5
    #speed = 15

    movement_event = game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    #movement_event = game.create_timer_trigger(20).on_reset(lambda x: bullet.move_across_dir((random.random()-0.5)*speed*0.3))


    frame_event = game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = game._create_specific_collision_trigger(bullet, game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        game.broadcast_message('player_health', -1)
        movement_event.remove()
        game.when_timer_reset(200, 1).add_handler(destory)

    when_exit_screen = game.when_condition_met(lambda: bullet.y > HEIGHT, repeats=1)

    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        game._remove_sprite(bullet)
        when_exit_screen.remove()

    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)






def create_enemy_type1(position):

    enemy_sprite = create_rect_sprite((255, 0, 0), 50, 30, pos=position)
    #game.add_sprite(enemy_sprite)

    enemy_sprite.add_rotation(90+(random.random()-0.5)*15)
    enemy_sprite.set_collision_type(ENEMY_TYPE)


    speed = random.random()*6




    movement_event = game.when_timer_reset(20).add_handler(lambda x: enemy_sprite.move_indir(speed))
    bullet_event = game.when_timer_reset(150).add_handler(lambda x: create_bullet_type1((enemy_sprite.x, enemy_sprite.y), enemy_sprite.get_rotation()))
    #bullet_event = game.create_timer_trigger(1500).on_reset(lambda x: create_bullet_attracted((enemy_sprite.x, enemy_sprite.y)))

    when_hit_player = game.when_condition_met(lambda: pyscratch.game.is_touching(game, enemy_sprite, game.shared_data['player']), repeats=1)
    when_leaving_screen = game.when_condition_met(lambda: (enemy_sprite.y > HEIGHT), repeats=1)
    when_hit_by_player_bullet = game._create_type2type_collision_trigger(PLAYER_BULLET_TYPE, ENEMY_TYPE)


    


    def destroy(x):
        when_hit_player.remove()
        when_leaving_screen.remove()
        when_hit_by_player_bullet.remove()
        movement_event.remove()
        bullet_event.remove()
        game._remove_sprite(enemy_sprite)

    def check_collision(a):
        if enemy_sprite._shape in a.shapes:
            destroy(None)

    when_hit_player.add_handler(destroy)
    when_hit_player.add_handler(lambda x: game.broadcast_message('player_health', -1))
    
    when_leaving_screen.add_handler(destroy)

    when_hit_by_player_bullet.add_handler(check_collision)






def game_start(data):

    player = create_rect_sprite((0, 0, 255), 50, 30, pos=(720//2, 1200))
    #game.add_sprite(player)
    create_edge_sprites(game)
    player.set_collision_type(PLAYER_TYPE)


    healthbar_red = create_rect((255, 0, 0), 60, 50)

    healthbar_empty = create_rect_sprite((255, 255, 255), 60, 5, pos=(0,0))



    game.shared_data['player'] = player
    game.shared_data['inaccuracy'] = 5

    player.sprite_data['health'] = 10


    #game.add_sprite(healthbar_empty)
    healthbar_empty.lock_to(player, (0,-30))
    healthbar_empty.blit(healthbar_red, (0,0))




    game.when_timer_reset(100, 20).add_handler(lambda x: create_bullet_move_sine((355, 1), 0))
    game.when_timer_reset(100, 20).add_handler(lambda x: create_bullet_attracted((355, 1)))
    game.when_timer_reset(100, 100).add_handler(lambda x: create_bullet_type1((100, 100), 90))

    #game.create_timer_trigger(100, 10)
    
    
    #game.create_timer_trigger(1200).on_reset(lambda x: shoot_player_bullet((player.x, player.y), game.shared_data['inaccuracy']))
    #game.create_timer_trigger(500, 30).on_reset(lambda x: create_enemy_type1((random.random()*WIDTH, 0)))


    def run_forever(_):
        if pyscratch.game.is_key_pressed('w'):
            player.move_xy((0, -5))

        if pyscratch.game.is_key_pressed('s'):
            player.move_xy((0, 5))

        if pyscratch.game.is_key_pressed('a'):
            player.move_xy((-5, 0))

        if pyscratch.game.is_key_pressed('d'):
            player.move_xy((5, 0))

        player.set_xy((cap(player.x, 50, WIDTH-50), cap(player.y, HEIGHT-500, HEIGHT)))

        
    def on_health_change(change):
        player.sprite_data['health'] += change
        new_health = max(0, player.sprite_data['health'])

        healthbar_red = create_rect((255, 0, 0), 60*(new_health/10), 50)
        healthbar_empty.blit(healthbar_red)



    game.when_timer_reset(1000/120).add_handler(run_forever)
    game.when_receive_message('player_health').add_handler(on_health_change)

wait_for_game_start = game.when_receive_message('game_start').add_handler(game_start)


game.start(60, 60, False, True)



from attr import s
import pyscratch as pysc
import pygame
from pyscratch import game

SCREEN_HEIGHT = 720
SCREEN_WIDTH = 1280

font = pygame.font.SysFont(None, 48)


score_board = pysc.create_rect_sprite((200, 200, 200), 150, 70, position=(SCREEN_WIDTH//2,SCREEN_HEIGHT//2))
game_start_event = score_board.when_game_start()
game['left_score'] = 0
game['right_score'] = 0

def display_score():
    score_board.show()
    l = game['left_score'] 
    r = game['right_score'] 
    score_board.write_text(f'{l} - {r}', font, offset=(150//2, 70//2))

game_start_event.add_handler(display_score)


left_score_event = score_board.when_receive_message('left_score')
def left_score(data):
    game['left_score'] += 1 
    display_score()

left_score_event.add_handler(left_score)


right_score_event = score_board.when_receive_message('right_score')
def right_score(data):
    game['right_score'] += 1 
    display_score()

right_score_event.add_handler(right_score)


def resume_game():
    game['running'] = True
    score_board.hide()

def on_space_release(updown):
    if  updown == 'up':
        resume_game()


game.when_key_pressed("space").add_handler(on_space_release)
score_board.when_this_sprite_clicked().add_handler(resume_game)





from utils.input_box import IntegerInputBox, FloatInputBoxHoriLabel
import pyscratch as pysc
from settings import *

xpos1 = 830
xpos2 = 1050

offset_x_input = IntegerInputBox('offset_x', 'offset_x_change', "0", 80)
offset_x_input.set_xy((xpos1, 605))


offset_y_input = IntegerInputBox('offset_y', 'offset_y_change', "0", 80)
offset_y_input.set_xy((xpos2, 605))

pysc.game['limit_x'] = None
pysc.game['limit_y'] = None

pysc.game['size_x'] = None
pysc.game['size_y'] = None



row_input = IntegerInputBox('n_row', "n_row_change", "", 80)
row_input.set_xy((xpos1, 650))

col_input = IntegerInputBox('n_col', "n_col_change", "", 80)
col_input.set_xy((xpos2, 650))


pixel_x_input = IntegerInputBox('pixel_x', 'pixel_x_change', "0", 80)
pixel_x_input.set_xy((xpos1, 695))

pixel_y_input = IntegerInputBox('pixel_y', 'pixel_y_change', "0", 80)
pixel_y_input.set_xy((xpos2, 695))


pysc.game.change_layer(offset_x_input, 1)
pysc.game.change_layer(offset_y_input, 1)
pysc.game.change_layer(row_input, 1)
pysc.game.change_layer(col_input, 1)
pysc.game.change_layer(pixel_x_input, 1)
pysc.game.change_layer(pixel_y_input, 1)



import pyscratch as pysc


player = pysc.create_single_costume_sprite("assets/player-fish.png")

# free to use any function name
def on_game_start():

    player.set_rotation_style_left_right() # analogous to the motion block: 'set rotation style [left-right]'
    while True:
        if pysc.is_key_pressed('w'): # analogous to the sensing block: 'key [w] pressed'
            player.y -= 4  # analogous to the motion block: change y by [-4]

        if pysc.is_key_pressed('s'):
            player.y += 4

        if pysc.is_key_pressed('a'):
            player.direction = 180  # analogous to the motion block: point in direction [180]
            player.x -= 4
            
        if pysc.is_key_pressed('d'):
            player.direction = 0
            player.x += 4


        # this is analogous to the control block: wait [1/60] seconds
        # because the frame rate is 60, this is basically to wait for one frame
        yield 1/60 

        # unlike scratch, the wait here is necessary. 
        # Without waiting here, python will put everything aside 
        # to attempt to run the loop as quickly as possible and thus 
        # halt everything else in the program.

        
# passing the function to the event as the event handler
game_start_event = player.when_game_start()
game_start_event.add_handler(on_game_start)

# or shorter: 
# player.when_game_start().add_handler(on_game_start)




from numpy import True_
import pyscratch as pysc
from pyscratch import game

SCREEN_HEIGHT = 720
SCREEN_WIDTH = 1280

# TODO: recommend a colour picking website
ball_colour = (220, 220, 220)
ball_radius = 25
ball_sprite = pysc.create_circle_sprite(ball_colour, ball_radius, position = (SCREEN_WIDTH//2, SCREEN_HEIGHT//2))
game['ball_sprite'] = ball_sprite

top_edge, left_edge, bottom_edge, right_edge = pysc.create_edge_sprites(thickness=0)

 
def movement():
    speed_x = 3
    speed_y = 3
    pong = False

    while True: 
        yield .5/game.framerate # wait for half a frame (for smoothness)
        

        if game['running']:
            
            ball_sprite.show()
            if ball_sprite.is_touching(top_edge):
                pong = True
                speed_y = abs(speed_y)

            elif ball_sprite.is_touching(bottom_edge):
                pong = True
                speed_y = -abs(speed_y)

            elif ball_sprite.is_touching(left_edge):
                speed_x = abs(speed_x)
                ball_sprite.broadcast_message('right_score')
                game['running'] = False
                pong = False

            elif ball_sprite.is_touching(right_edge):
                speed_x = -abs(speed_x)
                ball_sprite.broadcast_message('left_score')
                game['running'] = False
                pong = False

            elif ball_sprite.is_touching(game['right_paddle']):
                pong = True
                speed_x = -abs(speed_x)

            elif ball_sprite.is_touching(game['left_paddle']):
                pong = True
                speed_x = abs(speed_x)

            elif pong:
                game.play_sound('pong')
                pong = False

                
            ball_sprite.x += speed_x
            ball_sprite.y += speed_y
        else:
            ball_sprite.hide()
            
            ball_sprite.y = SCREEN_HEIGHT//2
            ball_sprite.x = SCREEN_WIDTH//2
    
game_start_event = ball_sprite.when_game_start()
game_start_event.add_handler(movement)




import pyscratch as pysc

spawn_button = pysc.create_single_costume_sprite("assets/button.png")

def spwan_enemy(): 
    enemy_red = pysc.create_single_costume_sprite("assets/fish_red_outline.png")
    enemy_blue = pysc.create_single_costume_sprite("assets/fish_blue_outline.png")
    
    enemy_red.x = 0
    enemy_red.y = pysc.game.screen_height/2


    enemy_blue.x = pysc.game.screen_width 
    enemy_blue.y = pysc.game.screen_height/2
    
    while True:
        enemy_red.x += 4
        enemy_blue.x -= 4

        yield 1/60  # must have an yield in a loop! 
    
spawn_button.when_this_sprite_clicked().add_handler(spwan_enemy) 




import pyscratch as pysc
game = pysc.game

chest = pysc.create_single_costume_sprite("assets/chatgpt-chest-open.png")

# 2. Basic Event
def game_start(): 
    chest.set_scale(0.15)
    chest.x = game.screen_width/2
    chest.y = game.screen_height/2
    game.move_to_back(chest)


chest.when_game_start().add_handler(game_start)

# 6. Backdrop
def game_end(): 
    chest.hide()

chest.when_backdrop_switched(1).add_handler(game_end)




from utils.input_box import IntegerInputBox, FloatInputBoxVerticallyLabel
import pyscratch as pysc
from settings import *
frame_itv = FloatInputBoxVerticallyLabel("frame_interval", default_value="0.2")

def set_xy():
    container = pysc.game['main_bottom_panel']
    frame_itv.lock_to(container, (0,0))
    frame_itv.set_xy((200, 30))
    
frame_itv.when_game_start().add_handler(
    set_xy
)

def new_switch(text0, text1, data_key):
    container = pysc.game.shared_data['main_bottom_panel']
    container_width, container_height = pysc.game.shared_data['main_bottom_size']

    button_w, button_h = 80, 50
    button = pysc.create_rect_sprite((221, 221, 221), button_w, button_h)
    
    pysc.game.shared_data[data_key] = False
    button.write_text(text0, DEFAULT_FONT48, colour=(0,0,0), offset=(button_w/2, button_h/2))

    def on_click():
        pysc.game.shared_data[data_key] = not pysc.game.shared_data[data_key]

        if pysc.game.shared_data[data_key]:
            button.write_text(text1, DEFAULT_FONT48, colour=(0,0,0), offset=(button_w/2, button_h/2))
        else:
            button.write_text(text0, DEFAULT_FONT48, colour=(0,0,0), offset=(button_w/2, button_h/2))
    
    button.when_this_sprite_clicked().add_handler(on_click)


    button.lock_to(container, (-(container_width-button_w)/2, -(container_height-button_h)/2), reset_xy=True)
    return button


def create_play_button():
    container = pysc.game.shared_data['main_bottom_panel']

    play_button = new_switch('Play','Pause', 'is_playing')
    play_button.lock_to(container, (0,0))
    
    play_button.set_xy((0, 70))

pysc.game.when_game_start().add_handler(create_play_button)





scale_factor_input = FloatInputBoxVerticallyLabel("scale_factor", "scale_factor_change", default_value="1")

def set_xy2():
    container = pysc.game.shared_data['main_bottom_panel']
    scale_factor_input.lock_to(container, (0,0))
    scale_factor_input.set_xy((200, 120))
    
scale_factor_input.when_game_start().add_handler(
    set_xy2
)




import pyscratch as pysc
import player

# start the game
screen_height = 720
screen_width = 1280
framerate = 60

pysc.game.update_screen_mode((screen_width, screen_height))
pysc.game.start(framerate)




import pyscratch as pysc
from pyscratch import game

def FishSprite():
    sp = pysc.create_animated_sprite("assets/some_fishes")

    def point_to_mouse():

        sp.x = pysc.random_number(0, game.screen_width)
        sp.y = pysc.random_number(0, game.screen_height)

        while True: 
            yield 1/game.framerate
            sp.point_towards_mouse()
            sp.move_indir(5)
        
    sp.when_game_start().add_handler(point_to_mouse)

    return sp

n_fish = 20
fishes = []
for i in range(n_fish):
    fishes.append(FishSprite())



from typing import List
import numpy as np

def get_dist(me: pysc.Sprite, others: List[pysc.Sprite]):

    distances = [me.distance_to_sprite(another) if not  me is another else np.inf for another in others ]
    ind = np.argsort(distances)#[::-1]

    ideal_dist = 200
    max_speed = 5

    div = ideal_dist/max_speed
    for n in ind[:n_fish-1]:
        n_nearest_other = others[n]

        if me.x > n_nearest_other.x:
            me.x += (ideal_dist - (me.x - n_nearest_other.x))/div

        elif me.x < n_nearest_other.x:
            me.x -= (ideal_dist - (n_nearest_other.x - me.x))/div

        if me.y > n_nearest_other.y:
            me.y += (ideal_dist - (me.y - n_nearest_other.y))/div

        elif me.y < n_nearest_other.y:
            me.y -= (ideal_dist - (n_nearest_other.y - me.y))/div


def keep_distance(_):
    for f in fishes:
        f.x = pysc.random_number(0, pysc.game.screen_width)
        f.y = pysc.random_number(0, pysc.game.screen_height)

    while True:
        yield 1/pysc.game.framerate
        for f in fishes:
            get_dist(f, fishes)

temp_e = pysc.game.when_timer_above(0)
temp_e.add_handler(keep_distance)


from utils.input_box import IntegerInputBox, FloatInputBoxVerticallyLabel
import pyscratch as pysc
from settings import *

button = pysc.create_rect_sprite((150, 150, 150), 220, 50)
button.write_text('Set current folder as sprite', DEFAULT_FONT24, offset=(110, 25))

def set_xy():
    container = pysc.game['main_bottom_panel']
    button.lock_to(container, (0,0))
    button.x = 0
    button.y = 0
    
button.when_game_start().add_handler(set_xy)


def on_click():
    pysc.game.broadcast_message('change_sprite_selection', pysc.game['path'])
    pass


button.when_this_sprite_clicked().add_handler(on_click)


from numpy import deg2rad
import pyscratch as pysc
import player, enemy, hearts

# backdrop
backdrop_img = pysc.load_image('assets/my_background.jpg') # load the image(s)
pysc.game.set_backdrops([backdrop_img]) # a list of all the backdrop images 
pysc.game.switch_backdrop(0) # use the backdrop at index 0 


# start the game
screen_height = 720
screen_width = 1280
framerate = 60

# add these to as shared variables so other sprites might use it
pysc.game['screen_height'] = screen_height
pysc.game['screen_width'] = screen_width
pysc.game['framerate'] = framerate

pysc.game.update_screen_mode((screen_width, screen_height))
pysc.game.start(framerate)




import pyscratch as pysc


player = pysc.create_single_costume_sprite("assets/player-fish.png")
pysc.game['player'] = player  # add the player as a shared variable so other sprites can access it


def on_game_start():

    player.set_rotation_style_left_right()
    speed_decay = 0.9
    speed_y = 0
    speed_x = 0

    while True:
        max_speed = 4


        if pysc.is_key_pressed('w'):
            speed_y = -max_speed

        elif pysc.is_key_pressed('s'):
            speed_y = max_speed

        else:
            speed_y *= speed_decay # speed_y = speed_y * speed_decay

        if pysc.is_key_pressed('a'):
            player.direction = 180
            speed_x = -max_speed
            
        elif pysc.is_key_pressed('d'):
            player.direction = 0
            speed_x = max_speed
        else:
            speed_x *= speed_decay # speed_x = speed_y * speed_decay


        player.y += speed_y
        player.x += speed_x

        yield 1/60


game_start_event = player.when_game_start()
game_start_event.add_handler(on_game_start)

# or shorter: player.when_game_start().add_handler(on_game_start)


def check_health():

    while True:
        if not pysc.game['health']:
            player.remove()
        yield 1/60


game_start_event.add_handler(check_health)



import pyscratch as pysc
from pyscratch import game


sp = pysc.create_animated_sprite("assets/used_by_examples/perspective_background/the_red_guy_piskel")  

def next_frame():
    while True:
        yield 0.2
        sp.next_frame()
sp.when_game_start().add_handler(next_frame)

def on_game_start():
    sp.x = 1280/2
    sp.y = 720/2
    sp.retrieve_saved_state()
    sp.set_scale(3)
    sp.set_draggable(True)

    sp.set_rotation_style_left_right()
    while True:
    
        yield 1/game.framerate

        is_running = False
        if pysc.is_key_pressed('d'):
            game['centre_x'] += 5
            is_running = True
            sp.direction=0
            
        if pysc.is_key_pressed('a'):
            game['centre_x'] -= 5
            is_running = True
            sp.direction=180

        if pysc.is_key_pressed('w'):
            game['centre_d'] += 5
            is_running = True
            

        if pysc.is_key_pressed('s'):
            game['centre_d'] -= 5
            is_running = True
        
        if is_running:
            sp.set_animation('run')
        else:
            sp.set_animation('idle')

        
            
sp.when_game_start().add_handler(on_game_start)


import pyscratch as pysc

# This is the line that create the sprite. Note that you need to match the image file name! 
player = pysc.create_single_costume_sprite("assets/fish_brown_outline.png")
player.set_draggable(True) # optional: make the sprite draggable



import pyscratch as pysc
from pyscratch import game
import chest, enemy, friend
       
# background
bg0 = pysc.load_image("assets/undersea_bg.png")
bg0 = pysc.scale_to_fit_aspect(bg0, (1024, 576))
game.add_backdrop('bg0', bg0)

bg1 = pysc.load_image("assets/Cat In Space Wallpaper Hq.jpg")
bg1 = pysc.scale_to_fit_aspect(bg1, (1024, 576))
game.add_backdrop('bg1', bg1)



def on_game_start():
    game.switch_backdrop('bg0')

game.when_game_start().add_handler(on_game_start)


# starting the game
game.update_screen_mode((1024, 576))
game.start(show_mouse_position=True)





import pyscratch as pysc
import random
from setting import *
import numpy as np



def create_bullet_attracted(position):
    bullet = pysc.Sprite(frames, "spin", position)
    
    bullet.set_scale(2.3)
    #pysc.game.add_sprite(bullet)
    speed = random.random()*15+5

    bullet.point_towards_sprite(pysc.game.shared_data['player'])


    movement_event = pysc.game.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    following_event = pysc.game.when_timer_reset(200, 5).add_handler(lambda x: bullet.point_towards_sprite(pysc.game.shared_data['player']))

    frame_event = pysc.game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = pysc.game._create_specific_collision_trigger(bullet, pysc.game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        pysc.game.broadcast_message('player_health', -1)
        movement_event.remove()
        pysc.game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        following_event.remove()
        hitting_player_event.remove()
        pysc.game._remove_sprite(bullet)

    when_exit_screen = pysc.game.when_condition_met(lambda: bullet.y > SCREEN_HEIGHT, repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)



def create_bullet_start_pointing(position, _):

    bullet = pysc.Sprite(frames, "spin", position)

    

    bullet.point_towards_sprite(pysc.game.shared_data['player'])
    bullet.set_scale(2.3)
    #pysc.game.add_sprite(bullet)
    speed = random.random()*15+5
    bullet.point_towards_sprite(pysc.game.shared_data['player'])


    movement_event = bullet.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    #movement_event = pysc.game.create_timer_trigger(20).on_reset(lambda x: bullet.move_across_dir((random.random()-0.5)*speed*0.3))


    frame_event = bullet.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = pysc.game._create_specific_collision_trigger(bullet, pysc.game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        pysc.game.broadcast_message('player_health', -1)
        movement_event.remove()
        pysc.game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        pysc.game._remove_sprite(bullet)

    when_exit_screen = pysc.game.when_condition_met(lambda: bullet.y > SCREEN_HEIGHT, repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)
    pass




def create_bullet_move_sine(position, rotation):
    bullet = pysc.Sprite(frames, "square_bullets", position)
    bullet.set_scale(2.3)
    #pysc.game.add_sprite(bullet)
    bullet.set_rotation(rotation+90)
    #speed = random.random()*15+5
    speed = 10


    bullet.sprite_data['phase'] = 0
    movement_event = pysc.game.when_timer_reset(30)
    def move(x):
        bullet.sprite_data['phase']+=.1
        angle = np.tanh(np.cos(.6*bullet.sprite_data['phase']))


        bullet.add_rotation(angle)
                            
        bullet.move_indir(speed)

    
    movement_event.add_handler(move)


    frame_event = pysc.game.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = pysc.game._create_specific_collision_trigger(bullet, pysc.game.shared_data['player'])

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        pysc.game.broadcast_message('player_health', -1)
        movement_event.remove()
        pysc.game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        movement_event.remove()
        frame_event.remove()
        hitting_player_event.remove()
        pysc.game._remove_sprite(bullet)

    when_exit_screen = pysc.game.when_condition_met(lambda: (bullet.y > SCREEN_HEIGHT) or (bullet.y < 0) or (bullet.x <0) or (bullet.x>SCREEN_WIDTH) , repeats=1)
    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)





def create_straight_bullet(position, rotation):
    """
    a bullet that moves a straight line
    """
    
    bullet = pysc.Sprite(frames, "spin", position)
    bullet.add_rotation(rotation)
    bullet.set_scale(2.3)
    #pysc.game.add_sprite(bullet)
    speed = random.random()*15+5
    #speed = 15

    movement_event = bullet.when_timer_reset(20).add_handler(lambda x: bullet.move_indir(speed))
    frame_event = bullet.when_timer_reset(200).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = bullet.create_specific_collision_trigger(pysc.game.shared_data['player'])
    when_exit_screen = bullet.when_condition_met(lambda: (bullet.y > SCREEN_HEIGHT) or (bullet.y < 0) or (bullet.x <0) or (bullet.x>SCREEN_WIDTH), repeats=1,)

    
    def explode_and_destroy(a):
        bullet.set_animation('star_explosion')
        pysc.game.broadcast_message('player_health', -1)
        movement_event.remove()
        pysc.game.when_timer_reset(200, 1).add_handler(destory)


    def destory(x):
        #movement_event.remove()
        #frame_event.remove()
        #hitting_player_event.remove()
        #when_exit_screen.remove()
        pysc.game._remove_sprite(bullet)

    when_exit_screen.add_handler(destory)
    hitting_player_event.add_handler(explode_and_destroy)





def create_exploding_bullet(position, rotation):
    """
    a bullet that explode after 1.2 second
    """

    bullet = pysc.Sprite(frames, "spin", position)
    bullet.add_rotation(rotation)
    bullet.set_scale(2.3)
    #pysc.game.add_sprite(bullet)

    speed = 7

    movement_event = pysc.game.when_timer_reset(20, associated_sprites=[bullet]).add_handler(lambda x: bullet.move_indir(speed))
    frame_event = pysc.game.when_timer_reset(200, associated_sprites=[bullet]).add_handler(lambda x: bullet.next_frame())
    hitting_player_event = pysc.game._create_specific_collision_trigger(bullet, pysc.game.shared_data['player'])
    exit_screen_event = pysc.game.when_condition_met(lambda: bullet.y > SCREEN_HEIGHT, repeats=1, associated_sprites=[bullet])
    explosion_event = pysc.game.when_timer_reset(1200,1, associated_sprites=[bullet])
        


    def hit_and_destroy(a):
        bullet.set_animation('star_explosion')
        movement_event.remove()
        pysc.game.when_timer_reset(200, 1).add_handler(destory)

    def spawn_sub_bullets(_):
        for i in range(12):
            create_straight_bullet(bullet._body.position, rotation=i*30)

    def delayed_spawn_sub_bullets(_):
        pysc.game.when_timer_reset(200, 1).add_handler(spawn_sub_bullets)

            
    def destory(x):
        

        #movement_event.remove()
        #frame_event.remove()
        #hitting_player_event.remove()
        #exit_screen_event.remove()
        #explosion_event.remove()
        pysc.game._remove_sprite(bullet)

    exit_screen_event.add_handler(destory)
    hitting_player_event.add_handler(hit_and_destroy)
    hitting_player_event.add_handler(lambda: pysc.game.broadcast_message('player_health', -1))
    explosion_event.add_handler(delayed_spawn_sub_bullets)
    explosion_event.add_handler(hit_and_destroy)




create_bullet_move_sine((355, 0), 90)






import pyscratch as pysc
from setting import *
import random 

def shoot_player_bullet(player):

    bullet = pysc.Sprite(frames, "circle_bullets", player.body.position)

    #pysc.game.add_sprite(bullet)
    bullet.set_collision_type(PLAYER_BULLET_TYPE)
    bullet.set_rotation(-90)

    movement_timer = pysc.game.when_timer_reset(1000/240).add_handler(
        lambda x: bullet.move_indir(2)
    )

    next_frame_timer = pysc.game.when_timer_reset(100).add_handler(
        lambda x: bullet.next_frame()
    )


    def destroy_when_exit(x):
        movement_timer.remove()
        next_frame_timer.remove()
        pysc.game._remove_sprite(bullet)


    destroy_condition = pysc.game.when_condition_met(lambda: (bullet.y < 0), repeats=1)
    destroy_condition.add_handler(destroy_when_exit)




pysc.game.when_receive_message('player_shoot_bullet').add_handler(shoot_player_bullet)



import pyscratch as pysc

player = pysc.create_single_costume_sprite("assets/fish_brown_outline.png")
player.set_draggable(True)

# 1. Create a function that does the thing (any function name is fine)
def rotate_move(): 

    player.direction += 45
    player.move_indir(20)
    yield 0.2 

    player.direction += 45
    player.move_indir(20)
    yield 0.2 

    player.direction += 45
    player.move_indir(20)
    yield 0.2 
    
    player.direction += 45
    player.move_indir(20)

# 2. Create an event object (any event name is fine)
click_event = player.when_this_sprite_clicked() 

# 3. Attach the function to the event block
click_event.add_handler(rotate_move) 

# Or step 2 and 3 together in one line
#player.when_this_sprite_clicked().add_handler(rotate_move) 



import pyscratch as pysc

player = pysc.create_single_costume_sprite("assets/fish_brown_outline.png")
player.set_draggable(True)

def scale(): 

    for i in range(20):
        player.scale_by(1.025)
        yield 0.03 

    for i in range(20):
        player.scale_by(0.975)
        yield 0.03 

player.when_this_sprite_clicked().add_handler(scale) 


# Remember: The function is the stack of scratch blocks without the event block at the top
def move(): 
    while True: # the forever loop 
        if pysc.is_key_pressed("d"): 
            player.x += 4  

        if pysc.is_key_pressed("a"):  
            player.x -= 4  

        if pysc.is_key_pressed("w"):  
            player.y -= 4  
            
        if pysc.is_key_pressed("s"):  
            player.y += 4  

        yield 1/60  # must have an yield in a loop! 

# Attach the function to the event
game_start = player.when_game_start() 
game_start.add_handler(move) 

# Or just in one line
#player.when_game_start().add_handler(move) 



import pyscratch as pysc
import pygame
import target, enemy, chest, friend#, text

game = pysc.game

# mouse 
pygame.mouse.set_visible(False)


# background
bg0 = pysc.load_image("assets/undersea_bg.png")
bg0 = pysc.scale_to_fit_aspect(bg0, (1024, 576))
game.add_backdrop('bg0', bg0)

bg1 = pysc.load_image("assets/Cat In Space Wallpaper Hq.jpg")
bg1 = pysc.scale_to_fit_aspect(bg1, (1024, 576))
game.add_backdrop('bg1', bg1)

game.switch_backdrop('bg0')

# sound
game.load_sound("hit", "assets/impactMetal_light_004.ogg")
game.load_sound("background", "assets/Circus-Theme-Entry-of-the-Gladiators-Ragtime-Version(chosic.com).mp3")

def play_loop():
    while True:
        game.play_sound("background", volume=0)
        yield 2*60+33
game.when_game_start().add_handler(play_loop)




def background():
    while True:
        yield 1/game.framerate
        if game['score'] < 0: 
            game.switch_backdrop('bg1')

        
game.when_game_start().add_handler(background)


game.update_screen_mode((1024, 576))
game.start(show_mouse_position=True)





import pyscratch as pysc
from pyscratch import game

chest = pysc.create_single_costume_sprite("assets/chatgpt-chest-open.png")

# 1. Create a function that does the thing (any function name is fine)
def set_size_position(): 
    """
    when game start:
    set the initial position of the chest
    """
    chest.set_scale(0.15)
    chest.x = game.screen_width/2
    chest.y = game.screen_height/2
    game.move_to_back(chest)

# 2. Create an event object (any event name is fine)
click_event = chest.when_this_sprite_clicked() 

# 3. Attach the function to the event block
click_event.add_handler(set_size_position) 

# Or step 2 and 3 together in one line
#chest.when_this_sprite_clicked().add_handler(set_size_position)


import pyscratch as pysc


player = pysc.create_single_costume_sprite("assets/player-fish.png")

# free to use any function name
def on_game_start():

    player.set_rotation_style_left_right()
    while True:
        if pysc.is_key_pressed('w'):
            player.y -= 4

        if pysc.is_key_pressed('s'):
            player.y += 4

        if pysc.is_key_pressed('a'):
            player.direction = 180
            player.x -= 4
            
        if pysc.is_key_pressed('d'):
            player.direction = 0
            player.x += 4

        yield 1/60 # because the frame rate is 60


game_start_event = player.when_game_start()
game_start_event.add_handler(on_game_start)

# or shorter: 
# player.when_game_start().add_handler(on_game_start)




import pyscratch as pysc
game = pysc.game

def spwan_enemy(data): 
    enemy_blue = pysc.create_single_costume_sprite("assets/fish_blue_outline.png")

    enemy_blue.y = pysc.random_number(0, game.screen_height)
    enemy_blue.x = 0

    player1 = game['player1']
    player2 = game['player2']
    
    # a better way: break the loop when the sprite is removed
    #while not enemy_blue.removed: 
    while True:
        enemy_blue.x += 4

        if enemy_blue.is_touching(player1):
            enemy_blue.remove()
            player1.scale_by(1.1)

        if enemy_blue.is_touching(player2):
            enemy_blue.remove()
            player2.scale_by(1.1)

        yield 1/game.framerate  # must have an yield in a loop! 
    
game.when_receive_message("spawn_enemy").add_handler(spwan_enemy) 

def on_start():
    while True:
        game.broadcast_message("spawn_enemy")
        yield 1

game.when_game_start().add_handler(on_start)



import pyscratch as pysc
from setting import *


def spawn_player(event_test=False):

   ...pyscratch# main sgame   player = pysc.create_rect_sprite((0, 0, 255), 50, 30, pos=(720//2, 1200))
    #pysc.game.add_sprite(player)
    player.set_collision_type(PLAYER_TYPE)


    # custom data
    pysc.game.shared_data['player'] = player
    pysc.game.shared_data['player_health'] = 10
    pysc.game.shared_data['bullet_period_ms'] = 200

    # behaviour
    ## 1. move by key press in a limited area
    ## 2. health changes on message
    ## 3. shoot bullet every n seconds (by messages)

    ## 1. movement    
    movement_event = pysc.game.when_timer_reset(1000/120)
    def movement(_):
        if game.is_key_pressed('w'):
            player.move_xy((0, -5))

        if game.is_key_pressed('s'):
            player.move_xy((0, 5))

        if game.is_key_pressed('a'):
            player.move_xy((-5, 0))

        if game.is_key_pressed('d'):
            player.move_xy((5, 0))

        player.set_xy((cap(player.x, 50, SCREEN_WIDTH-50), cap(player.y, SCREEN_HEIGHT-900, SCREEN_HEIGHT)))
    
    movement_event.add_handler(movement)


    ## 2. health changes on message
    health_change_event = pysc.game.when_receive_message('player_health')

    def health_change(change):
        pysc.game.shared_data['player_health'] += change

    health_change_event.add_handler(health_change)


    ## 3. bullets 
    bullet_timer = pysc.Timer()
    condition = lambda: (bullet_timer.read() > pysc.game.shared_data['bullet_period_ms'])
    shoot_bullet_event = pysc.game.when_condition_met(condition)

    def shoot_bullet(n):
        pysc.game.broadcast_message('player_shoot_bullet', player)
        bullet_timer.full_reset()

    shoot_bullet_event.add_handler(shoot_bullet)

    

    # health bar
    healthbar_empty = pysc.create_rect_sprite((255, 255, 255), 60, 5, pos=(0,0))
    #pysc.game.add_sprite(healthbar_empty)
    healthbar_empty.lock_to(player, (0,-30))

    ## move the health bar
    healthbar_red = pysc.create_rect((255, 0, 0), 60, 50)
    healthbar_empty.blit(healthbar_red, (0,0))

    # reusing the health_change_event defined about
    def on_health_change(change):
        new_health = max(0, pysc.game.shared_data['player_health'])

        healthbar_red = pysc.create_rect((255, 0, 0), 60*(new_health/10), 50)
        healthbar_empty.blit(healthbar_red)
    
    health_change_event.add_handler(on_health_change)

    if event_test: 
        shoot_bullet_event.remove()




spawn_player(event_test=False)




import pyscratch as pysc
from pyscratch import game

enemy = pysc.create_single_costume_sprite("assets/fish_red_skeleton_outline.png")


def clicked():
    """
    when the enemy is clicked: 
    change the enemy location with a fade out effect
    """
    game['score'] += 10


    for i in range(10):
        yield 1/game.framerate
        enemy.set_transparency(1-i/10) # print(1-i/10) if you find this confusing.
        enemy.scale_by(0.9)

    enemy.x = pysc.random_number(0, game.screen_width)
    enemy.y = pysc.random_number(0, game.screen_height)

    enemy.set_transparency(1)
    enemy.set_scale(1)
    
enemy.when_this_sprite_clicked().add_handler(clicked)


def movement():
    """
    when_game_start: 
    the movement of the enemy
    """

    centre = (game.screen_width/2, game.screen_height/2)

    while True:
        yield 1/game.framerate

        speed = 5+game['score']*0.1
        speed = max(1, speed)


        # get the distance to the mouse
        mouse_x, mouse_y = pysc.get_mouse_pos()
        distance_to_mouse = enemy.distance_to((mouse_x, mouse_y))

        # enemy avoids the mouse when it is close to the mouse
        if distance_to_mouse < 200:
            enemy.point_towards_mouse()
            enemy.direction += 180
            enemy.move_indir(speed)
        
        # otherwise, go to the centre of the screen 
        else:
            enemy.point_towards(centre)
            enemy.move_indir(speed)


enemy.when_game_start().add_handler(movement)



def reduce_score():
    """
    reduce the score for every 0.1 second the enemy is in the centre
    """
    centre = (game.screen_width/2, game.screen_height/2)

    while True:
        yield 0.1
        
        if enemy.distance_to(centre) < 50:
            game['score'] -= 1

enemy.when_game_start().add_handler(reduce_score)
